const webpack = require('webpack');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const ESLintPlugin = require('eslint-webpack-plugin');
const ProgressBarPlugin = require('progress-bar-webpack-plugin');
const getCSSModuleLocalIdent = require('./getCSSModuleLocalIdent');

const sassRegex = /\.(scss|sass)$/;
const sassModuleRegex = /\.module\.(scss|sass)$/;
const imageNotResolvingRegex = /^\/(common-uiasset|wwwasset|onlineasset)\//;
const httpRegex = /^https?:\/\//;

function getStyleLoaders({ isProd, isServer, isHotServer, isSass, moduleOption = {} }) {
  const result = [
    {
      loader: require.resolve('css-loader'),
      options: {
        ...moduleOption,
        url: {
          // background image不解析以"http、https"开头或者/common-uiasset/|/wwwasset/|/onlineasset/开头的静态资源， 例如：'/common-uiasset/img.png'
          filter: url => !httpRegex.test(url) && !imageNotResolvingRegex.test(url)
        }
      }
    },
    {
      loader: require.resolve('postcss-loader'),
      options: {
        postcssOptions: {
          ident: 'postcss',
          config: false,
          plugins: [
            'postcss-flexbugs-fixes',
            [
              'postcss-preset-env',
              {
                autoprefixer: {
                  flexbox: 'no-2009'
                },
                stage: 3
              }
            ],
            require('postcss-normalize').default
          ]
        },
        sourceMap: false
      }
    }
  ];

  if (isSass) {
    result.push(
      {
        loader: require.resolve('resolve-url-loader')
      },
      {
        loader: require.resolve('sass-loader'),
        options: {
          // 增加环境变量，供sass使用
          additionalData: `$env: ${isProd ? 'production' : 'development'};`
        }
      }
    );
  }

  if ((isProd || isHotServer) && !isServer) {
    result.unshift({
      loader: MiniCssExtractPlugin.loader
    });
  }

  if (!isProd && !isHotServer) {
    result.unshift({
      loader: require.resolve('style-loader')
    });
  }

  return result;
}

module.exports = (appConfig, extendEnv, { isProd, isServer, isHotServer }) => {
  const result = {
    stats: 'errors-only',
    // 原始源代码（仅限行）定位到原始代码问题，比source-map 快，但是没有列映射(column mapping)的 source-map
    devtool: isProd ? 'source-map' : 'cheap-module-source-map',
    // 会将 process.env.NODE_ENV 的值设为 development, 告知 webpack 使用相应模式的内置优化
    mode: isProd ? 'production' : 'development',
    // 打包的出口
    output: {
      publicPath: '/'
    },
    cache: {
      type: 'filesystem',
      version: `${appConfig.projectName}_${isProd ? '1' : '0'}${isServer ? '1' : '0'}${isHotServer ? '1' : '0'}_${
        appConfig.projectVersion
      }`,
      store: 'pack',
      buildDependencies: {
        defaultWebpack: ['webpack/lib/'],
        config: [__filename]
      }
    },
    infrastructureLogging: {
      level: 'none'
    },
    // 配置各种loader
    module: {
      // 警告替换为Error
      strictExportPresence: true,
      rules: [
        {
          oneOf: [
            {
              test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
              type: 'asset',
              parser: {
                dataUrlCondition: {
                  maxSize: 2048
                }
              },
              generator: {
                filename: `static/[name].[contenthash:8].[ext]?v=${extendEnv.version}`
              }
            },
            {
              test: /\.(js|ts|tsx)$/,
              exclude: /node_modules/,
              use: {
                loader: require.resolve('babel-loader'),
                options: {
                  cacheDirectory: true,
                  presets: [
                    [
                      require('@babel/preset-env').default,
                      {
                        useBuiltIns: 'entry',
                        corejs: 3,
                        exclude: ['transform-typeof-symbol']
                      }
                    ],
                    require('@babel/preset-react').default,
                    require('@babel/preset-typescript').default
                  ],
                  plugins: [
                    require('@loadable/babel-plugin').default,
                    require('@babel/plugin-transform-runtime').default,
                    [require('@babel/plugin-proposal-decorators').default, { legacy: true }],

                    [
                      require('@babel/plugin-proposal-class-properties').default,
                      {
                        loose: true
                      }
                    ],
                    [
                      require('@babel/plugin-proposal-private-methods').default,
                      {
                        loose: true
                      }
                    ],
                    [
                      require('@babel/plugin-proposal-private-property-in-object').default,
                      {
                        loose: true
                      }
                    ],
                    // 支持 "??"
                    require('@babel/plugin-proposal-nullish-coalescing-operator').default,
                    // 支持"?."
                    require('@babel/plugin-proposal-optional-chaining').default
                  ]
                }
              }
            },
            {
              test: sassRegex,
              exclude: sassModuleRegex,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer,
                isSass: true
              })
            },
            {
              test: sassModuleRegex,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer,
                isSass: true,
                modulesOptions: {
                  importLoaders: 3,
                  sourceMap: false,
                  modules: {
                    mode: 'local',
                    getLocalIdent: getCSSModuleLocalIdent
                  }
                }
              })
            },
            {
              test: /\.(css)$/,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer
              })
            }
          ]
        }
      ]
    },
    resolve: {
      // 文件后缀自动补全
      extensions: ['.tsx', '.ts', '.js', 'scss', 'css']
    },
    plugins: [
      new webpack.DefinePlugin(extendEnv.stringified),
      new ESLintPlugin({
        exclude: 'node_modules',
        threads: 4,
        failOnError: false,
        lintDirtyModulesOnly: isProd,
        extensions: ['tsx', 'ts', 'js']
      }),
      new ProgressBarPlugin({
        // eslint-disable-next-line no-useless-concat
        format: 'build [:bar] ' + ':percent' + ' (:elapsed seconds)',
        clear: true
      })
    ]
  };

  if (isProd) {
    // build下, 忽略locale moment, 缩小代码体积
    result.plugins.push(
      new webpack.IgnorePlugin({
        resourceRegExp: /^\.\/locale$/,
        contextRegExp: /moment$/
      })
    );
  }

  if (isProd && !isServer) {
    result.optimization = {
      chunkIds: 'named',
      minimize: true
    };
  }

  return result;
};
